// JUMA-PA100 protocol
// Juha Niinikoski, OH2NLT 06.07.2006

// Active only if TRX2 protocol is selected 15.11.2008
/*
 The PA-100D Protocol uses the following commands:
 Query Band
 The query command from the Juma TRX-2 is: =?B\n\r
 The PA-100D will respond: =Bn\n\r
 Where n can take the value from 0 - 10
 0 = Out-Of-Band
 1 = 160m
 2 = 80m
 3 = 40m
 4 = 30m
 5 = 20m
 6 = 17m
 7 = 15m
 8 = 12m
 9 = 10m
 10 = Unknown

 Set Band
 The set band command from the Juma TRX-2 is: =Bn\n\r
 Where n has the same meanings as for the query response.
*/

#include <stdio.h>
#include <ctype.h>
#include "juma-pa100.h"				// Board definitions
#include "pa100_eeprom.h"			// Get EEPROM structure definitions

// External References
// External Functions
extern void putch(char);			// UART I/O
extern unsigned char kbhit(void);
extern unsigned char getch(void);

// Global variables
int poll_resp_rec;					// Poll response (freq set) received from TRX-2/KX3, 1 = Received
const char Send_Data[] = {"=B%i\n\r"};

// Module variables
unsigned char cmd_buf[MAX_BUFFER];	// Command buffer
int cmd_buf_idx = 0;				// Buffer fill pointer

// EEPROM Structures
// User Configuration Data
extern union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval) / 2];
	} eeprom;

// System Calibration Values
extern union
	{
	struct calval calval;
	unsigned int ee[sizeof(struct calval) / 2];
	} cal;

// TRX2 protocol parser
// Clear buffer
// This method uses less code and is more efficient than a FOR-NEXT loop by 8 instructions! It clears the buffer, and leaves
// cmd_buf_idx reset to zero when the procedure exits. A.Ryan - 5B4AIY
void clear_buffer(void)
	{
	cmd_buf_idx = MAX_BUFFER;

	do {cmd_buf[--cmd_buf_idx] = 0x00;} while (cmd_buf_idx);
	}

// Parse TRX2 message
void serial_pa100(void)
	{
	char c;

	if(kbhit())
		{
		c = getch();						// Get next character

//		printf("\n\rCharacter: %c $%2.2X, Index: %i", c, c, cmd_buf_idx);		// test
		if(c == 0x0A) return;				// Ignore Newline

		switch(cmd_buf_idx)
			{
			case 0:							// Possible start of message
				if(c != '=') return;		// If it is not start of message character, ignore.
			break;

			case 1:							// If second character is not a query or a command,
				if(c != '?' && c != 'B')
					{
					cmd_buf_idx = 0;		// then reset index, and return.
					return;
					}
			break;

			case 2:							// If the third character is invalid,
				if((cmd_buf[1] == '?' && c != 'B') || (cmd_buf[1] == 'B' && !isdigit(c)))
					{
					cmd_buf_idx = 0;		// then reset index, and return.
					return;
					}
			break;

			case 3: default:
				if(c == 0x0D)				// If fourth character is end-of-message byte,
					{
					if(cmd_buf[1] == 'B')	// and the second is 'B', then this is a set band command,
						{
						eeprom.defval.band = cmd_buf[2] - '0';
						poll_resp_rec = TRUE;	// so set the new band, and the polling response flag.
						}
					else					// Otherwise, it is a send band command to the TRX-2.
						printf(Send_Data, (eeprom.defval.band == NOT_KNOWN) ? 0 : eeprom.defval.band);
					}						// If it is not the end-of-message byte,
				clear_buffer();				// and in any event, clear the buffer and restart.
				return;
			break;
			}
		cmd_buf[cmd_buf_idx++] = c;			// Add the character to the buffer.
		}
	}

